import json

try:
    from django.contrib.postgres.fields import JSONField
    from django.contrib.postgres.forms.jsonb import JSONField as JSONFormField
except ImportError:
    from django.db.models import JSONField
    from django.forms import JSONField as JSONFormField

__all__ = ('UTF8JSONFormField', 'UTF8JSONField')


class UTF8JSONFormField(JSONFormField):
    def prepare_value(self, value):
        try:
            return json.dumps(value, ensure_ascii=False)
        except json.JSONDecodeError:
            return value


class UTF8JSONField(JSONField):
    """JSONField for postgres databases.

    Displays UTF-8 characters directly in the admin, i.e. äöü instead of
    unicode escape sequences.
    """

    def formfield(self, **kwargs):
        return super().formfield(**{**{'form_class': UTF8JSONFormField}, **kwargs,})



from django.db.models import Manager
from dynamicfield_serializer import DynamicFieldSerializer
from rest_framework.serializers import ListSerializer, LIST_SERIALIZER_KWARGS

from .utils import replacer, calculate, get_operations

__all__ = (
    'CalculationDynamicFieldSerializer',
    'GeneratorSerializer',
    'GeneratorListSerializer',
)


class GeneratorListSerializer(ListSerializer):
    def to_representation(self, data):
        iterable = data if isinstance(data, Manager) else data
        for item in iterable:
            yield self.child.to_representation(item)

    @property
    def data(self):
        return list(self.to_representation(self.instance))

    @property
    def generator(self):
        return self.to_representation(self.instance)


class GeneratorSerializer(DynamicFieldSerializer):
    """if many=True, have property generator - to return generator serializable values"""

    @classmethod
    def many_init(cls, *args, **kwargs):
        allow_empty = kwargs.pop('allow_empty', None)
        child_serializer = cls(*args, **kwargs)
        list_kwargs = {
            'child': child_serializer,
        }
        if allow_empty is not None:
            list_kwargs['allow_empty'] = allow_empty
        list_kwargs.update(
            {
                key: value
                for key, value in kwargs.items()
                if key in LIST_SERIALIZER_KWARGS
            }
        )
        return GeneratorListSerializer(*args, **list_kwargs)


class CalculationDynamicFieldSerializer(GeneratorSerializer):
    def to_representation(self, instance):
        ret = super().to_representation(instance)
        calc = self.context.get("calculation")
        if calc:
            calc = get_operations(calc)
            for name, calc_value in calc.items():
                if name in ret:
                    data = replacer(calc_value, name, ret[name])
                    ret[name] = calculate(data)

        return ret



from rest_framework.exceptions import APIException


class InvalidDateParams(APIException):
    status_code = 422
    default_code = 'invalid_date'
    default_detail = 'Invalid date params.'


class ReportUpdating(APIException):
    status_code = 204


class DontHaveReportData(APIException):
    status_code = 406
    default_detail = 'Dont have reports for this integration'


class InvalidFieldsParam(APIException):
    status_code = 422
    default_code = 'invalid_fields'
    default_detail = 'Invalid fields param, fields must be a list'



import requests
import importlib
from urllib.parse import unquote

from rest_framework.exceptions import PermissionDenied
from django.conf import settings


__all__ = (
    'CredentialMixin',
    'SmartAdminRequiredMixin',
)


class BaseCredentialMixin(object):

    if settings.DASHBOARD_URL.endswith("/"):
        url = settings.DASHBOARD_URL
    else:
        url = settings.DASHBOARD_URL + '/'

    def get_user_info(self, request):
        if 'token' in request.GET and "platform_id" in request.GET:
            token = unquote(request.GET["token"])
        elif 'HTTP_AUTHORIZATION' in request.META and "platform_id" in request.GET:
            token = request.META["HTTP_AUTHORIZATION"]
        else:
            raise PermissionDenied(
                {"status": "error", "message": "you don't have permission to access"}
            )
        try:
            response = requests.get(
                f"{self.url}api/check-platform/?platform_id={request.GET['platform_id']}",
                headers={"executetoken": token},
            )
        except requests.ConnectionError:
            raise PermissionDenied(
                {"status": "error", "message": "Authorization failed."}
            )

        if response.status_code != 200:
            raise PermissionDenied(
                {"status": "error", "message": "Authorization failed."}
            )

        return response.json()["auth_info"]

    def get_check_admin(self, request, *args, **kwargs):
        user_info = self.get_user_info(request, *args, **kwargs)
        user_id = user_info["id"]
        if user_id in (19, 22, 23, 55):
            return user_id
        raise PermissionDenied(
            {"status": "error", "message": "This endpoint only for smart admin users."}
        )


class CredentialMixin(BaseCredentialMixin):

    check_credential = True
    with_smart_auth = True

    # method for dashboard app
    def get_check_dash_auth(self, request):
        if 'HTTP_EXECUTETOKEN' not in request.META:
            raise PermissionDenied(
                {"status": "error", "message": "you don't have permission to access"}
            )
        auth_info = requests.post(
            f"{self.url}api/check-auth/",
            headers={"executetoken": request.META['HTTP_EXECUTETOKEN']},
        ).json()
        if "is_admin" not in auth_info:
            raise PermissionDenied({"status": "error", "message": "Invalid dash user."})

        return auth_info["is_admin"]

    def check_mongo_credential(self):
        mongo_credential = getattr(settings, 'MONGODANTIC_CREDENTIAL', False)
        return mongo_credential

    def get_credential(self, request, model=None, mongodantic=False):
        """[summary]

        Arguments:
            request {HttpRequest} -- django http request

        Keyword Arguments:
            model {django model or None} -- django model (default: {None})

        Raises:
            PermissionDenied: if not credential exists

        Returns:
            django model -- credential object
        """
        if not model:
            model = self.get_credential_model()

        cred_id = request.GET.get("credential_id")
        mongodantic = mongodantic or self.check_mongo_credential()
        if not mongodantic:
            params = {
                'id': cred_id,
            }

            credential = model.objects.filter(**params).first()
        else:
            params = {
                '_id': cred_id,
            }

            credential = model.querybuilder.find_one(**params)
        if not credential:
            raise PermissionDenied(
                {"status": "error", "message": "Credential does not exist."}
            )

        return credential

    def initial(self, request, *args, **kwargs):
        if self.with_smart_auth:
            user_info = self.get_user_info(request)
            self.smart_user_id = user_info['id']
            self.smart_username = user_info['username']
            if hasattr(settings, 'CREDENTIAL_MODEL') and self.check_credential:
                self.credential = self.get_credential(request)
        return super().initial(request, *args, **kwargs)

    def get_credential_model(self):
        if hasattr(settings, 'CREDENTIAL_MODEL'):
            module_path, model = settings.CREDENTIAL_MODEL.rsplit('.', 1)
            module = importlib.import_module(module_path)
            return getattr(module, model)
        raise AttributeError('miss CREDENTIAL_MODEL in django settings')


class SmartAdminRequiredMixin(BaseCredentialMixin):
    def initial(self, request, *args, **kwargs):
        self.get_check_admin(request)
        return super().initial(request, *args, **kwargs)



from typing import Optional
from django.http import HttpResponseRedirect
from rest_framework.views import APIView
from rest_framework.viewsets import ModelViewSet
from rest_framework.response import Response
from rest_framework.exceptions import PermissionDenied
from rest_framework.views import APIView

from .exceptions import InvalidDateParams, InvalidFieldsParam
from .mixins import CredentialMixin
from .pagination import LimitOffsetPaginationListAPIView

__all__ = (
    'HomeAPIView',
    'BaseCredentialModelViewSet',
    'BaseReportListAPIView',
    'BaseGetCredentialAPIView',
    'BaseStatTSVAPIView',
)


class BaseCredentialModelViewSet(CredentialMixin, ModelViewSet):
    check_credential = False
    queryset = None
    serializer_class = None

    def get_queryset(self):
        if not self.queryset:
            raise NotImplementedError("model must be declare")
        return self.queryset

    def create(self, request, format=None):
        raise NotImplementedError()

    def retrieve(self, request, *args, **kwargs):
        """
        ---
        parameters:
            - Authorization token in headers
        """

        _ = self.get_check_dash_auth(request)
        return super().retrieve(request, *args, **kwargs)

    def list(self, request, *args, **kwargs):
        """
        ---
        parameters:
            - Authorization token in headers
        """
        _ = self.get_check_dash_auth(request)

        return super().list(request, *args, **kwargs)


# base View model
class BaseReportListAPIView(LimitOffsetPaginationListAPIView):
    serializer_class = None
    model = None
    select_related_fields = ()  # must be tuple or list
    with_platforms_sample = False
    order_by_fields = ()  # must be tuple or list
    filter_params = {}  # custom filter params

    def get_model(self):
        if not self.model:
            raise NotImplementedError("model must be declare")
        return self.model

    def get_queryset(self):
        if self.select_related_fields and isinstance(
            self.select_related_fields, (tuple, list)
        ):
            qs = (
                self.get_model()
                .objects.select_related(*self.select_related_fields)
                .filter(
                    integration_id=self.integration_id,
                    date__range=[self.date_from, self.date_to],
                    **self.filter_params,
                )
            )
        else:
            qs = self.get_model().objects.filter(
                integration_id=self.integration_id,
                date__range=[self.date_from, self.date_to],
                **self.filter_params,
            )
        if self.with_platforms_sample:
            return qs.filter(platform_id=self.platform_id)
        if self.order_by_fields:
            return qs.order_by(*self.order_by_fields)
        return qs


class BaseGetCredentialAPIView(CredentialMixin, APIView):
    def get_redirect_uri(self):
        """Create in this method redirect uri
            Example: https://oauth.yandex.ru/authorize?response_type=code&client_id={settings.YANDEX_APP_CLIENT_ID}&state={self.get_state()}&redirect_uri={redirect_url}
        """
        raise NotImplementedError()

    def check_auth_params(self):
        """This method for check auth params"""
        if (
            "callback_url" not in self.request.GET
            or "token" not in self.request.GET
            or "platform_id" not in self.request.GET
        ):
            raise PermissionDenied(
                {"status": "error", "message": "You dont have permission to access."}
            )
        return self.request

    def get_state(self):
        """
            Create state in this method
            Excample:
                user_info = self.get_user_info(request, get_token=True)
                user_id = user_info["id"]
                main_user = user_info["username"]
                # user_id = 16
                # main_user = "main_user"

                state = {
                    "callback_url": request.GET['callback_url'],
                    "user_id": user_id,
                    "main_user": main_user,
                    "platform_id": request.GET['platform_id']
                }
                return state
        """
        raise NotImplemented()

    def get(self, request, format=None):
        """
        ---
        parameters:
            - Authorization token in headers
            - 'callback_url' in GET params
        """

        return HttpResponseRedirect(self.get_redirect_uri())


class HomeAPIView(CredentialMixin, APIView):
    def get(self, request, format=None):
        return Response({"status": "success", "message": "smart integration api"})


class BaseStatTSVAPIView(CredentialMixin, APIView):
    check_credential = False
    serializer_class = None

    def get_serializer_class(self):
        if not self.serializer_class:
            raise NotImplementedError('serializer_class must be implemented!')
        return self.serializer_class

    def check_log(self, *args, **kwargs) -> None:
        """check any download log,

        Returns:
            None: or raise APIException
        """
        return None

    def validate_request_data(self, request) -> dict:
        admin = self.get_check_admin(request)
        # return Response({"status": "error", "message": "Dont have statistic for this integration"}, status=400)
        if not admin:
            raise PermissionDenied(
                {
                    "status": "error",
                    "message": "You don't have permissions for this endpoint.",
                }
            )
        data = request.data
        integration_id = data.get("integration_id")

        if not integration_id:
            return (
                InvalidDateParams(
                    {"status": "error", "message": "Miss integration id."}, status=422
                ),
            )

        values = list()
        if 'fields' in data:
            serializer = self.get_serializer_class()
            fields = list(serializer().fields.keys())
            for f in data['fields']:
                if f in fields:
                    values.append(f)
                else:
                    raise InvalidFieldsParam(
                        {"status": "error", "message": f"invalid field - {f}"}
                    )
        self.check_log()
        return {
            'data': data,
            'integration_id': integration_id,
            'admin': admin,
            'platform_id': request.GET['platform_id'],
            'values': values,
        }




from datetime import datetime, timedelta
from rest_framework.pagination import LimitOffsetPagination
from rest_framework.generics import ListAPIView
from rest_framework.response import Response
from rest_framework.exceptions import NotAcceptable

from .mixins import CredentialMixin

__all__ = ("LimitOffsetPaginationListAPIView",)


class StandardResultsPagination(LimitOffsetPagination):
    default_limit = 1000
    max_limit = 1000


class LimitOffsetPaginationListAPIView(CredentialMixin, ListAPIView):
    """
    limit=1000
    offset=1000
    for calculation:
        rewrite get_calculation method -> dict
    for dynamic fields:
        (Need DynamicFieldSerializer)
        rewrite get_fields  method -> list
    you always have:
    self.integration_id
    self.date_form
    self.date_to
    """

    check_credential = False
    limit = None
    offset = None
    pagination_class = StandardResultsPagination

    def _ensure_limit_param(self):
        if self.limit:
            self.pagination_class.default_limit = self.limit

    def _generate_date_params(self, delta: int) -> tuple:
        date_to = datetime.today().strftime("%Y-%m-%d")
        date_from = (datetime.today() - timedelta(days=delta)).strftime("%Y-%m-%d")
        return date_from, date_to

    def _get_params(self):
        self.integration_id = self.request.GET.get("integration_id")
        last_week = self.request.GET.get("last_week", "")
        last_3_days = self.request.GET.get("last_3_days", "")
        date_from = self.request.GET.get("date_from", "")
        date_to = self.request.GET.get("date_to", "")

        if not self.integration_id:
            raise NotAcceptable({"status": "error", "message": "miss integration_id"})

        if last_week == "true":
            date_from, date_to = self._generate_date_params(7)
        elif last_3_days == "true":
            date_from, date_to = self._generate_date_params(3)

        if not date_from and not date_to:
            raise NotAcceptable({"status": "error", "message": "miss date params"})
        self.date_from = date_from
        self.date_to = date_to
        self.platform_id = self.request.GET["platform_id"]

    def get_calculation(self) -> dict:
        return {}

    def get_fields(self) -> list:
        return []

    def list(self, request, format=None):
        self._get_params()
        self._ensure_limit_param()
        queryset = self.get_queryset()
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(
                page,
                many=True,
                context={
                    "calculation": self.get_calculation(),
                    "response_fields": self.get_fields(),
                },
            )
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(
            queryset,
            many=True,
            context={
                "calculation": self.get_calculation(),
                "response_fields": self.get_fields(),
            },
        )

        return Response(serializer.data)



import csv
from hashlib import md5
from datetime import datetime, timedelta
from typing import Optional, Any, Union, Generator
from rest_framework.response import Response

from .exceptions import InvalidDateParams
from .tools import (
    get_file_link,
    check_file_exist,
    upload_to_s3,
    delete_file,
    has_items,
)


class BaseDataManager(object):
    model: Any = None
    serializer: Any = None
    default_limit: int = 150000
    query_limit: int = 1000
    with_last_id: bool = True
    last_id_field: str = 'id'

    def __init__(self, integration_id: int, date_from: str, date_to: str):
        self.integration_id = integration_id
        self.date_from = date_from
        self.date_to = date_to

    def get_serializer_class(self) -> Any:
        if not self.serializer:
            raise NotImplementedError(
                'get_serializer_class or serializer must be implemented.'
            )
        return self.serializer

    def get_model(self) -> Any:
        if not self.model:
            raise NotImplementedError('get_model or model must be implemented.')
        return self.model

    def get_count(self) -> int:
        raise NotImplementedError('get_count must be implemented.')

    def get_data(self, skip: int = 0, **kwargs) -> list:
        raise NotImplementedError('get_data must be implemented.')

    def get_default_serializer_fields(self) -> list:
        serializer = self.get_serializer_class()
        fields = list(serializer().fields.keys())
        return fields

    def serialize_data(self, data: Any, fields: list) -> Union[Generator, list, dict]:
        raise NotImplementedError('serialize_data must be implemented.')

    def _generate_file(self, s3_filename: str, fields: list, skip: int):
        with open(s3_filename, 'w') as f:
            writer = csv.DictWriter(f, delimiter='\t', fieldnames=fields)
            writer.writeheader()
            last_id = None
            while True:
                result = self.get_data(skip=skip, last_id=last_id)
                exist, data = has_items(result)
                if not exist:
                    break
                if self.with_last_id:
                    last_id = (
                        getattr(data[-1], self.last_id_field)
                        if isinstance(data, list)
                        else getattr(list(data)[-1], self.last_id_field)
                    )
                data = self.serialize_data(data, fields)
                for row in data:
                    writer.writerow(row)
                skip += self.query_limit

    def _generate_file_link(self, skip: int, s3_filename: str, fields: list) -> tuple:
        if check_file_exist(s3_filename):
            skip += self.default_limit
            return get_file_link(s3_filename)
        self._generate_file(s3_filename, fields, skip)
        s3_file_data = upload_to_s3(s3_filename)
        return skip, get_file_link(s3_file_data['s3_filename'])

    def _generate_result(self, filename: str, fields: list) -> dict:
        skip = 0
        part = 0
        count = self.get_count()
        links = []
        if self.default_limit >= count:
            skip, link = self._generate_file_link(skip, f'{filename}.tsv', fields)
            links.append(link)
            delete_file(f'{filename}.tsv')
        else:
            while skip <= count:
                while skip <= self.default_limit:
                    part += 1
                    part_filename = f'{filename}_{part}'
                    s3_filename = f"{part_filename}.tsv"
                    skip, link = self._generate_file_link(
                        skip, f'{filename}.tsv', fields
                    )
                    links.append(link)
                    delete_file(s3_filename)
        return {'links': links}

    def generate(self, filename: str, fields: Optional[list] = None) -> dict:
        fields: list = fields if fields else self.get_default_serializer_fields()
        return self._generate_result(filename=filename, fields=fields)


class BaseStatisticSelector(object):
    selector_type: str = 'base'

    def __init__(
        self,
        data: dict,
        integration_id: int,
        admin: int,
        platform_id: str = 'prod',
        values: list = [],
    ):
        self.data = data
        self.integration_id = integration_id
        self.admin = admin
        self.platform_id = platform_id
        self.values = values

    def get_date_params(self) -> any:
        date_from = None
        date_to = None
        if self.data.get("last_week", "").lower() == 'true':
            date_to = datetime.today().strftime('%Y-%m-%d')
            last_date = datetime.today() - timedelta(days=7)
            date_from = last_date.strftime('%Y-%m-%d')
        elif self.data.get("last_3_days", "").lower() == 'true':
            date_to = datetime.today().strftime('%Y-%m-%d')
            last_date = datetime.today() - timedelta(days=3)
            date_from = last_date.strftime('%Y-%m-%d')

        elif self.data.get('date_from') and self.data.get('date_to'):
            date_to = self.data['date_to']
            date_from = self.data['date_from']

        if not date_to or not date_from:
            raise InvalidDateParams(
                {
                    "status": "error",
                    "message": "Miss date param (like last_week,last_3_days, date_from and date_to)",
                }
            )
        return date_from, date_to

    def generate_statistic_response(
        self, date_from: str, date_to: str, filename: str
    ) -> Response:
        raise NotImplementedError()

    def generate_file_name(self, date_from: str, date_to: str) -> str:
        if self.values:
            hash_ = md5(str(self.values).encode('utf-8')).hexdigest()
            filename = f'/tmp/{self.selector_type}_{self.admin}_{self.integration_id}_{hash_}{date_from}:{date_to}'
        else:
            filename = f'/tmp/{self.selector_type}_{self.admin}_{self.integration_id}_{date_from}:{date_to}'
        return filename

    def response(self) -> Response:
        date_from, date_to = self.get_date_params()
        filename = self.generate_file_name(date_from=date_from, date_to=date_to)
        return self.generate_statistic_response(
            date_from=date_from, date_to=date_to, filename=filename,
        )

